/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLCNF:					*/
/*	Take care of all clement configuration	*/
/*	structure reading from the data-base.	*/
/*						*/
/************************************************/
#include	<string.h>
#include	<errno.h>
#include	<netdb.h>
#include	<arpa/inet.h>
#include	<ctype.h>

#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"unidns.h"
#include	"unisql.h"
#include	"unijrl.h"
#include	"gesdbc.h"
#include	"gescnf.h"
#include	"lvlcnf.h"
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a config need	*/
/*	to be reloaded				*/
/*						*/
/************************************************/
static time_t loadtime(CNFTYP *cnf,SQLFILE *sql)

{
time_t freshdate;
time_t lstupd;
int phase;
int proceed;

freshdate=(time_t)0;
lstupd=freshdate;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*DB config date	*/
      if (cnf!=(CNFTYP *)0)
        lstupd=rly_lastupd(&(cnf->relais));
      break;
    case 1	:	/*DB config date	*/
      lstupd=dbc_checkbind(sql);
      break;
    case 2	:	/*DB relais config	*/
      lstupd=dbc_checkrelay(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  if (lstupd>freshdate)
    freshdate=lstupd;
  phase++;
  }
return freshdate;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to update ban/trust database	*/
/*	with a list of updated IP		*/
/*						*/
/************************************************/
int cnf_updbantrust(SQLFILE *sql,char **lists,int status,char *origin)

{
#define OPUB    "lvlcnf.c:dbupdate,"

register int num;
register RLYTYP *relais;
register int phase;
register int proceed;

num=0;
relais=(RLYTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch(phase) {
    case 0	:	/*getting relay file	*/
      relais=cnf_loadrlyfile(sql,relais);
      break;
    case 1	:	/*lists against relay	*/
      if ((lists!=(char **)0)&&(relais!=(RLYTYP *)0)) {
	time_t curtime;
	char **nlists;
	int i;

	curtime=time((long *)0);
	nlists=(char **)0;
	for (i=0;lists[i]!=(char *)0;i++) {
	   if (rly_isinrelay(relais,lists[i],curtime)==false)
	     nlists=(char **)vec_addlstlst((void **)nlists,(void *)lists[i]);
	   else {
	     lists[i]=clm_freestr(lists[i]);
	     }
	   }
	if (nlists!=(char **)0) {
	  (void) memcpy(lists,nlists,sizeof(char *)*(vec_sizelstlst((void **)nlists)+1));
	  (void) free(nlists);
	  }
	}
      break;
    case 2	:	/*lists against private	*/
      if ((lists!=(char **)0)&&(lists[0]!=(char *)0)) {
	static char *priv[]={"192.168.0.1/16","10.0.0.0/8","172.16.0.0/12",(char *)0};

	register int i;

	for (i=0;(lists!=(char **)0)&&(priv[i])!=(char *)0;i++) {
	  int mask;
	  char cleanip[30];
	  AFNTYP *privafn;

	  (void) sscanf(priv[i],"%49[^/]/%d",cleanip,&mask);
	  if ((privafn=afn_getipnum(cleanip))!=(AFNTYP *)0) {
	    char **nlists;
	    int j;

	    nlists=(char **)0;
	    (void) afn_maskip(privafn,mask);
	    for (j=0;lists[j]!=(char *)0;j++) {
	      AFNTYP *afnnum;

	      if ((afnnum=afn_getipnum(lists[j]))==(AFNTYP *)0) {
		(void) clm_alert(0,"%s Unable to scan IP <%s> (error=<%s>)",OPUB,lists[j],strerror(errno));
		lists[j]=clm_freestr(lists[j]);
		}
	      else {
		if (afn_cmpipnum(privafn,afnnum,privafn->afnmask)==false)
		  nlists=(char **)vec_addlstlst((void **)nlists,(void *)lists[j]);
		else {
		  lists[j]=clm_freestr(lists[j]);
		  }
		afnnum=afn_freeipnum(afnnum);
		}
	      }
	    if (nlists!=(char **)0) {
	      (void) memcpy(lists,nlists,sizeof(char *)*(vec_sizelstlst((void **)nlists)+1));
	      (void) free(nlists);
	      }
	    privafn=afn_freeipnum(privafn);
	    }
	  }
	}
      break;
    case 3	:	/*updating data base	*/
      if ((lists!=(char **)0)&&(lists[0]!=(char *)0)) 
	switch (status) {
	  case RBANNED	:
  	    num=dbc_setban(sql,lists,status,origin);
	    break;
	  case TRSTED	:
	    if (lists!=(char **)0) {/*always	*/
	      register int i;

	      for (i=0;lists[i]!=(char *)0;i++) {
		if (dbc_settrust(sql,lists[i],TRSTRS,true,time((time_t *)0))==true)
		  num++;
	        }
	      }
	    break;
	  default	:
	    (void) clm_alert(0,"%s Unexpected status value='%d' (bug?)",OPUB,status);
	    break;
          }
      break;
    case 4	:	/*cleaning relay file	*/
      relais=rly_freerelais(relais);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return num;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to load the relayed IP	*/
/*	found in configuration			*/
/*						*/
/************************************************/
RLYTYP *cnf_loadrlyfile(SQLFILE *sql,RLYTYP *rly)

{
RLYTYP *relais;
char **strrly;

relais=rly_loadrlyfile(rly);
strrly=(char **)0;
if ((strrly=dbc_getrelay(sql))!=(char **)0) {
  relais=rly_addrelais(relais,strrly[0],strrly+1);
  strrly=(char **)vec_freelstlst((void **)strrly,VECFREE);
  }
return relais;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to build a config structure	*/
/*						*/
/************************************************/
CNFTYP *cnf_chkconfig(SQLFILE *sql,CNFTYP *cnf)

{
/*checking database update every UPDCYCLE sec	*/
#ifdef	DEBUG
#define	UPDCYLE	5	
#else
#define	UPDCYLE	60	
#endif

time_t curtime;
time_t lstupd;
time_t update;
int phase;
int proceed;

curtime=time((time_t *)0);
lstupd=(time_t)0;
update=(time_t)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*config already loaded?*/
      if (cnf!=(CNFTYP *)0) {
	cnf->updated=false;
	update=cnf->update;
	if ((cnf->lastscan+UPDCYLE)>curtime) {
	  proceed=false;/*no need to recheck	*/
	  }
        }
      break;
    case  1	:	/*was a binding update?	*/
      lstupd=loadtime(cnf,sql);
      if (lstupd<=update)
	phase=999;	/*no need to go further	*/
      break;
    case  2	:	/*reloading config	*/
      cnf=cnf_freeconfig(cnf);
      cnf=(CNFTYP *)calloc(1,sizeof(CNFTYP));
      break;
    case  3	:	/*loading the new config*/
      cnf->binds=dbc_getbinds(sql,(BININF **)0,pro_smtp);
      cnf->binds=dbc_getbinds(sql,cnf->binds,pro_smtps);
      break;
    case  4	:	/*loading relais data	*/
      cnf->relais=cnf_loadrlyfile(sql,cnf->relais);
      break;
    case  5	:	/*updating update	*/
      if (cnf->relais!=(RLYTYP *)0) {
	time_t updrelay;

	updrelay=rly_lastupd(&(cnf->relais));
	if (updrelay>lstupd)
	  lstupd=updrelay;
	}
      cnf->updated=true;
      cnf->lastscan=curtime;
      cnf->update=lstupd;
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return cnf;
}
