/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	MODTRT:					*/
/*	Define E-mail TRansiT module		*/
/*						*/
/************************************************/
#include	<sys/poll.h>
#include	<sys/stat.h>
#include	<errno.h>
#include	<time.h>
#include	<math.h>
#include	<string.h>
#include	<unistd.h>

#include	"lowtyp.h"
#include	"utlprc.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"unijrl.h"
#include	"unisql.h"
#include	"geseml.h"
#include	"gestru.h"
#include	"gesque.h"
#include	"gesclm.h"
#include	"geseml.h"
#include	"gesdbc.h"
#include	"lvleml.h"
#include	"lvltrt.h"
#include	"modtrt.h"

#define	TITLE	"%s Transitr (%04lu/%08lu)"

#define	NOREPLY	"No_reply@localhost.localdomain"

/*how much to delay email if remote server busy	*/
#define	POSTDLY	60
#define	TRYWARM	 3	/*advice when trouble to*/
			/*send email to foreign	*/

typedef	struct	{
	pid_t chldnum;	/*process child number	*/
	char *recdom;	/*domain served		*/
	}SRVTYP;

typedef	enum	{
	srt_dest,	/*check for destination	*/
	srt_mx,		/*check for MX data	*/
	srt_unknown	/*buggy check		*/
	}SRTTYP;

/*Default string to open data-base		*/
static	const	char *empty=(char *)0;
/*

*/
/************************************************/
/*						*/
/*	Procedure to set status value		*/
/*	within known value			*/
/*						*/
/************************************************/
static int setstatvalue(int status)

{
if ((status>=CMDOK)&&(status<(CMDOK+10)))
  status= CMDOK;
if ((status>=TMPERR)&&(status<SYNERR))
  status=RETRY;
if (status>=SYNERR)
  status=FAILED;
switch (status) {
  case CMDOK	:
  case RETRY	:
  case FAILED	:
    break;	/*nothing to do			*/
  default	:
    (void) clm_alert(0,"modtrt.c:setstatvalue, Unexpected status value '%d' (bug?)",status);
    status=RETRY;
    break;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a specific pid	*/
/*	is within the list of running process	*/
/*						*/
/************************************************/
static int stillup(PRCTYP *prclst,pid_t chldnum)

{
int isup;

isup=false;
if (prclst!=(PRCTYP *)0) {
  unsigned int i;

  for (i=0;i<prclst->nbr;i++) {
    if (prclst->pqueue[i]==chldnum) {
      isup=true;
      break;
      }
    }
  }
return isup;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add a domain and a process	*/
/*	pid to a working process list		*/
/*						*/
/************************************************/
static SRVTYP **addtosrvlst(SRVTYP **srvlst,char *recdom,pid_t chldnum)

{
SRVTYP *srvloc;

srvloc=(SRVTYP *)calloc(1,sizeof(SRVTYP));
srvloc->chldnum=chldnum;
srvloc->recdom=strdup(recdom);
srvlst=(SRVTYP **)vec_addlstlst((void **)srvlst,(void *)srvloc);
return srvlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if remote smtp server*/
/*	is still served.			*/
/*						*/
/************************************************/
static SRVTYP **chksrvlst(PRCTYP *prclst,SRVTYP **srvlst)

{
if (srvlst!=(SRVTYP **)0) {
  int i;

  for (i=0;srvlst[i]!=(SRVTYP *)0;i++) {
    int j;

    if (stillup(prclst,srvlst[i]->chldnum)==true)
      continue;
    (void) free(srvlst[i]->recdom);
    (void) free(srvlst[i]);
    srvlst[i]=(SRVTYP *)0;
    j=i;
    do
      {
      j++;
      srvlst[j-1]=srvlst[j];
      }
    while (srvlst[j]!=(SRVTYP *)0);
    i--; 
    }
  if (srvlst[0]==(SRVTYP *)0) {
    (void) free(srvlst);
    srvlst=(SRVTYP **)0;
    }
  }
return srvlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get a clean MX list NOT	*/
/*	including our own IP (in case we are	*/
/*	relaying Email within our own domain to	*/
/*	a non responding server, then calling us*/
/*	as secondary MX)			*/
/*						*/
/************************************************/
static MXTYP **getcleanmxlst(BININF **binlst,char *recdom)

{
MXTYP **mxlst;

if ((mxlst=dns_getmxorip(recdom))!=(MXTYP **)0) {
  register int i;

  for (i=0;mxlst[i]!=(MXTYP *)0;i++) {
    int found;
    char **iplist;

    found=false;
    /*do not remove IP if host is the first MX for domain	*/
    /*we are not in the case to be relaying Emailas secondary MX*/
    if (i==0)	
      continue; 
    if ((iplist=dns_getiplist(mxlst[i]->mxname))!=(char **)0) {
      register int j;

      for (j=0;(iplist[j]!=(char *)0)&&(found==false);j++) {
	register int k;

        for (k=0;binlst[k]!=(BININF *)0;k++) {
	  if (strcmp(iplist[j],binlst[k]->ip)==0) {
	    found=true;
	    break;
	    }
	  }
	}
      iplist=(char **)vec_freelstlst((void **)iplist,VECFREE);
      }
    if (found==true) {
      register int j;

      j=i;
      mxlst[j]=dns_cleanonemx(mxlst[j]);
      do {
	mxlst[j]=mxlst[j+1];
	j++;
	}
      while (mxlst[j]!=(MXTYP *)0);
      i--;
      }
    }
  if (mxlst[0]==(MXTYP *)0) {
    (void) free(mxlst);
    mxlst=(MXTYP **)0;
    }
  }
return mxlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if a remote server	*/
/*	for a specific server is too busy	*/
/*	still up and running			*/
/*						*/
/************************************************/
static int istoobusy(SRVTYP **srvlst,char *recdom,int max)

{
int toobusy;
int num;

toobusy=false;
num=0;
if (srvlst!=(SRVTYP **)0) {
  int i;

  for (i=0;srvlst[i]!=(SRVTYP *)0;i++) {
    if (strcmp(srvlst[i]->recdom,recdom)==0)
      num++;
    }
  }
if (num>=max) 
  toobusy=true;
return toobusy;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if max child are	*/
/*	still up and running			*/
/*						*/
/************************************************/
static int allbusy(PRCTYP **prclst,SRVTYP ***srvlst,unsigned int max)

{
int busy;

busy=false;
*prclst=prc_childcheck(*prclst);
*srvlst=chksrvlst(*prclst,*srvlst);
if ((*prclst!=(PRCTYP *)0)&&((*prclst)->nbr>=max)) 
  busy=true;
return busy;
}
/*

*/
/************************************************/
/*	Procedure to report something within	*/
/*	tracefile.				*/
/*						*/
/************************************************/
static void logrpt(QUETYP *queue,const char *fmt,...)

{
if (queue!=(QUETYP *)0) {
  va_list args;
  char *strloc;
  char *rpt;
  time_t curtime;
  char logtime[30];

  va_start(args,fmt);
  (void) vasprintf(&strloc,fmt,args);
  curtime=time((long *)0);
  (void) strftime(logtime,sizeof(logtime),"%H:%M:%S",localtime(&curtime));
  (void) asprintf(&rpt,"%s %s",logtime,strloc);
  queue->logs=(char **)vec_addlstlst((void *)queue->logs,(void *)rpt);
  (void) free(strloc);
  va_end(args);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create a transit tracelog	*/
/*						*/
/************************************************/
static void tracelog(QUETYP *queue,char **traces) 

{
if (traces!=(char **)0) {
  int i;

  for (i=0;traces[i]!=(char *)0;i++) {
    queue->logs=(char **)vec_addlstlst((void *)queue->logs,(void *)strdup(traces[i]));
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to log queue entry status	*/
/*	destination.				*/
/*						*/
/************************************************/
static void logqueue(SQLFILE *sql,QUETYP *queue) 

{
char *recipient;

(void) asprintf(&recipient,"%s@%s",queue->recname,queue->recdom);
(void) dbc_updlogs(sql,queue->session,recipient,eml_getcomstat(queue->status,(MENTYP)queue->orgstatus));
(void) free(recipient);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to sort queue according	*/
/*	destination.				*/
/*						*/
/************************************************/
static QUETYP **sortqueue(QUETYP **queues,SRTTYP srtmode) 

{
QUETYP **dest;

dest=(QUETYP **)0;
if ((queues!=(QUETYP **)0)&&(queues[0]!=(QUETYP *)0)) {
  register int qulong;
  register int i;

  qulong=vec_sizelstlst((void **)queues);
  dest=(QUETYP **)calloc(qulong+1,sizeof(QUETYP *));
  dest[0]=queues[0];
  for (i=1;i<qulong;i++) {
    register int notequal;

    notequal=true;
    switch (srtmode) {
      case srt_dest	:
        notequal=strcmp(dest[0]->recdom,queues[i]->recdom)!=0;
	break;
      case srt_mx	:
        notequal=strcmp(dest[0]->smtpmx,queues[i]->smtpmx)!=0;
	break;
      default		:
	(void) clm_alert(0,"modtrt.c:sortqueue, Unexpected sort mode '%d' (bug?)",srtmode);
        break;
      }
    if (notequal)
      break;
    dest[i]=queues[i];
    }
  qulong++;
  (void) memmove(queues,queues+i,(qulong-i)*sizeof(QUETYP *));
  }
return dest;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to set the transit status	*/
/*	explainantion.				*/
/*	allow Email transit.			*/
/*						*/
/************************************************/
static void setwhy(QUETYP **queues,int status,char *frm, ...)

{
/*ATTENTION Note: bug within vasprintf!		*/
/*		  va_list is lost, do not call	*/
/*		  it twice with the same args!	*/
va_list args;
char *strloc;

va_start(args,frm);
(void) vasprintf(&strloc,frm,args);
(void) jrl_multijrl(EVTRT,"",strloc);
if (queues!=(QUETYP **)0) {
  int i;

  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    queues[i]->status=status;
    queues[i]->why=clm_freestr(queues[i]->why);
    queues[i]->why=strdup(strloc);
    }
  }
(void) free(strloc);
va_end(args);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to to postpone email to be	*/
/*	sent for a shot time.			*/
/*						*/
/************************************************/
static void postpone(QUETYP **queues)

{
#define	OPPO	"modtrt.c:postpone,"

SQLFILE *sql;
int phase;
int proceed;

sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*warning message	*/
      (void) setwhy(queues,FAILED,"Domain <%s> feeder process too busy (rescheduling '%d' recipients)",
				   queues[0]->recdom,vec_sizelstlst((void **)queues));
      break;
    case 1	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to re-open database",OPPO);
	phase=999;	/*big big trouble	*/
	}
      break;
    case 2	:	/*saving courante queue	*/
      if (queues!=(QUETYP **)0){	/*always*/
	int i;
	time_t nexttime;

	nexttime=time((time_t *)0)+60;
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  queues[i]->progress=QWAIT;
	  queues[i]->nexttime=nexttime;
          (void) dbc_updqueue(sql,queues[i]);
	  }
	}
      break;
    case 3	:	/*new scheduling	*/
      (void) dbc_postponequeue(sql,queues[0]->recdom,POSTDLY);
      break;
    case 4	:	/*closing data-base	*/
      sql=sql_close(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to establish a contact with	*/
/*	one of the available MX			*/
/*						*/
/************************************************/
static CTTYP *makecontact(MXTYP **mxlst,QUETYP *queue,time_t sectimeout)

{
CTTYP *contact;

contact=(CTTYP *)0;
(void) logrpt(queue,"#-- Ready to try %d MX connection",vec_sizelstlst((void **)mxlst));
if (mxlst!=(MXTYP **)0) {
  register int i;

  for (i=0;mxlst[i]!=(MXTYP *)0;i++) {
    int socker;

    socker=0;
    (void) logrpt(queue,"#-- Trying <%s> port '%s'",mxlst[i]->mxname,queue->smtpport);
    contact=eml_opencontact(pro_smtp,mxlst[i]->mxname,queue->smtpport,sectimeout,&socker);
    if (contact!=(CTTYP *)0) {
      if (contact->established==true) { 
        (void) jrl_journalling(false,EVTRT,"Connected to <%s/[%s]> port '%s'",
					    contact->peername,contact->peerip,
					    contact->servport);
        (void) logrpt(queue,"#-- Connected to <%s/[%s]> port '%s'",
					    contact->peername,contact->peerip,
					    contact->servport);
        break;
        }
      if (contact->info!=(char *)0) {
        (void) logrpt(queue,"#-- contact not established got <%s>",contact->info);
        (void) jrl_journalling(false,EVTRT,"%s",contact->info);
	}
      contact=eml_closecontact(contact);
      }
    else {
      char *strloc;

      (void) asprintf(&strloc,"Unable to make SMTP contact with <%s> port '%s' (error=<%s>)",
			       mxlst[i]->mxname,queue->smtpport,strerror(socker));
      (void) jrl_journalling(false,EVTRT,"%s",strloc);
      (void) logrpt(queue,"#-- %s",strloc);
      (void) free(strloc);
      }
    }
  }
return contact;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to know about the email to be	*/
/*	send size.				*/
/*						*/
/************************************************/
static u_long getmailsize(QUETYP *queue)

{
u_long msize;
char *datafile;
struct stat mailstat;

msize=(u_long)0;
datafile=que_getdatafilename(d_transit,queue);
if (stat(datafile,&mailstat)<0) 
  (void) clm_alert(0,"%s Unable to get file <%s> stat infomation (error=<%s> bug?)",
		     "modtrt.c:getmailsize",datafile,strerror(errno));
else
  msize=(u_long)mailstat.st_size;
(void) free(datafile);
return msize;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send Email contents on	*/
/*	connected SMTP channel.			*/
/*						*/
/************************************************/
static int senddata(CTTYP *contact,QUETYP *queue)

{
int status;
FILE *fdata;
char *datafile;
int phase;
int proceed;

status=RETRY;
fdata=(FILE *)0;
datafile=que_getdatafilename(d_transit,queue);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((fdata=fopen(datafile,"r"))==(FILE *)0) {
	status=FAILED;
	contact->info=clm_freestr(contact->info);
	(void) asprintf(&(contact->info),"unable to open file <%s> (error=<%s>)",datafile,strerror(errno));
	proceed=false;	/*Big Trouble		*/
	}
      break;
    case 1	:	/*is remote ready to get*/
      if (eml_sendcmd(contact,DATAOK,"%s",DATA)!=DATAOK) 
	phase=999;	/*get out fast		*/
      break;
    case 2	:	/*sending file contents		*/
      status=eml_contactstream(contact,fdata);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      if (status!=CMDOK) 
	phase=999;	/*trouble trouble		*/
      break;
    case 3	:	/*asking acknowledge	*/
      status=eml_sendcmd(contact,CMDOK,"%s",DATAEND);
      (void) jrl_multijrl(EVTRT,"",contact->info);
      break;
    default	:	/*SAFE guard		*/
      (void) fclose(fdata);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(datafile);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send queued email on SMTP	*/
/*	channel					*/
/*						*/
/************************************************/
static void sendemail(SQLFILE *sql,QUETYP **q,CTTYP *contact)

{
int nxt;
int last;
int batch;

nxt=0;
for (last=0;q[last]!=(QUETYP *)0;last++);
batch=nxt;
while (q[nxt]!=(QUETYP *)0) {
  char *session;
  u_long msize;
  int fromstat;
  int num;

  session=q[nxt]->session;
  msize=getmailsize(q[nxt]);
  fromstat=ERRPROC;
  num=0;
  /*NO NEED TO SEND a RSET if more than one mail to be sent	*/
  (void) jrl_journalling(false,EVTRT,"%s %s",SESSION,session);
  (void) jrl_journalling(false,EVTRT,"From: <%s>",q[nxt]->orgeml);
  (void) jrl_journalling(false,EVTRT,"Subject: %s",q[nxt]->subject);
  fromstat=eml_sendmfrom(contact,q[nxt]->orgeml,msize);
  (void) clm_cleanstring(contact->info);
  while (q[nxt]!=(QUETYP *)0) {
    register int status;

    status=fromstat;
    if (strcmp(q[nxt]->session,session)!=0) {
      break; 
      }
    if (status==CMDOK) {
      char *strloc;

      (void) asprintf(&strloc,"%s@%s",q[nxt]->recname,q[nxt]->recdom);
      (void) jrl_journalling(false,EVTRT,"%s %05lu",TRANSIT,q[nxt]->transit);
      (void) jrl_journalling(false,EVTRT,"To: <%s>",strloc);
      status=eml_sendcmd(contact,CMDOK,"%s: <%s>",RCPT,strloc);
      status=setstatvalue(status);
      switch (status) {
	case CMDOK	:
          num++;
	  break;
        default		:
          contact->info=clm_cleanstring(contact->info);
          (void) jrl_multijrl(EVTRT,"\t",contact->info);
	  q[nxt]->why=clm_freestr(q[nxt]->why);
	  (void) asprintf(&(q[nxt]->why),"Received from <%s>\n\t%s",contact->peername,contact->info);
	  break;
	}
      (void) free(strloc);
      }
    else 
      q[nxt]->why=strdup(contact->info);
    q[nxt]->status=status;
    nxt++;
    }
  if (num>0) {
    register int status;
    register int i;
    char *strloc;

    strloc=(char *)0;
    status=senddata(contact,q[batch]);
    (void) clm_cleanstring(contact->info);
    switch (status) {
      case ERRPROC	:
      case NOANSWR	:
        (void) asprintf(&strloc,"Status=%03d, error with <%s>\n\t\t%s",
				 status,contact->peername,contact->info);
    	contact->established=false;
	nxt=last;	/*NOT responding, no need to go further	*/
	break;
      case CMDOK	:
      default		:
        (void) asprintf(&strloc,"Received from <%s>\n\t\t%s",contact->peername,contact->info);
        status=setstatvalue(status);
	break;
      }
    (void) jrl_journalling(false,EVTRT,"%s",strloc);
    for (i=batch;i<nxt;i++) {
      if (q[i]->status==CMDOK) {
        q[i]->status=status;
	q[i]->why=clm_freestr(q[i]->why);
	q[i]->why=strdup(strloc);
	}
      }
    (void) free(strloc);
    }
  if (q[nxt]!=(QUETYP *)0) {
  /*sending a RSET there is another Email		*/
    if (eml_sendcmd(contact,CMDOK,"%s",RSET)!=CMDOK) {
      int i;

      for (i=nxt;q[i]!=(QUETYP *)0;i++) {
        q[i]->status=INRELAY;
        q[i]->why=clm_freestr(q[i]->why);
        (void) asprintf(&(q[i]->why),"Received RSET error from <%s>\n%s",
				      contact->peername,clm_cleanstring(contact->info));
        }
      nxt=i;		/*No need to go further	*/
      }
    }
  if ((q[nxt]==(QUETYP *)0)&&(contact->established==true)) {
    /*no Email to send anymore			*/
    (void) eml_sendcmd(contact,QUITOK,SQUIT);
    contact->established=false;
    }
  if (q[batch]!=(QUETYP *)0) {	/*always	*/
    char **traces;
    int i;

    traces=tls_gettraces(contact->tls);
    for (i=batch;i<nxt;i++) {
      (void) tracelog(q[i],traces);
      (void) logqueue(sql,q[i]);
      }
    traces=(char **)vec_freelstlst((void **)traces,VECFREE);
    }
  batch=nxt;
  } 
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to quarantine an E-mail in the*/
/*	user directory within the right domain	*/
/*	Return result within the Queue element.	*/
/*						*/
/************************************************/
static void quaremail(QUETYP *q)

{
#define OPST	"modtrt.c:quaremail,"

int done;
char *newfile;
char *mailfile;
char *newdir;
int phase;
int proceed;

done=false;
(void) asprintf(&newdir,"%s/%s/%s/%s",clm_appdir(d_mails),q->recdom,q->recname,QUARDIR);
(void) asprintf(&newfile,"%s/%s",newdir,q->session);
mailfile=que_getdatafilename(d_transit,q);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*need to create newdir	*/
      (void) clm_createdirs(newdir,0755);
      break;
    case 1	:	/*moving  filename	*/
      done=true;
      if (link(mailfile,newfile)<0) {
	done=false;		/*oups trouble		*/
	switch (errno) {
	  case EXDEV	:
	    done=clm_dupfile(mailfile,newfile);
	    break;
	  default	:
	    q->status=CMDOK;
            (void) asprintf(&(q->why),"unable to move file <%s> to <%s> (error=<%s>)",
				       mailfile,newfile,strerror(errno));
	    break;
	  }
	}
      if (done==true) {
	q->status=CMDOK;
        q->why=clm_freestr(q->why);
        (void) asprintf(&(q->why),"mail now available in <%s@%s> quarantine box",q->recname,q->recdom);
	}
      break;
    case 2	:	/*set quar file mode	*/
      if (chmod(newfile,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)<0) 
	(void) clm_alert(0,"%s unable to change file <%s> permission mode (error=<%s>)",
			    OPST,newfile,strerror(errno));
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(mailfile);
(void) free(newfile);
(void) free(newdir);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do check if recipient	*/
/*	domain need to be overridden		*/
/*						*/
/************************************************/
static void checkdomover(SQLFILE *sql,QUETYP *queue)

{
char *dom;

if ((dom=dbc_isitadomainalias(sql,queue->recdom))!=(char *)0) {
  (void) jrl_journalling(false,EVTRT,"Domain override from <%s@%s> to <%s@%s>",
				      queue->recname,queue->recdom,queue->recname,dom);
  (void) free(queue->recdom);
  queue->recdom=dom;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transmit Email to the	*/
/*	designated PIPE.			*/
/*						*/
/************************************************/
static void piping(QUETYP *q)

{
#define OPPI	"modtrt.c:piping,"

char *mailfile;
char *exec;
int phase;
int proceed;

mailfile=que_getdatafilename(d_transit,q);
exec=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*checking PIPE command	*/
      if (q->recname[0]!='|') {
	(void) clm_alert(0,"%s Unexpected pipe syntaxe <%s> (bug?)",OPPI,q->recname);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 1	:	/*making pipe string	*/
      (void) memmove(q->recname,q->recname+1,strlen(q->recname));
      (void) clm_skipblank(q->recname);
      if (q->recname[0]!='/')
        (void) asprintf(&exec,"%s%s/%s",clm_appdir(d_usrlib),SHELL,q->recname);
      else
        (void) asprintf(&exec,"%s",q->recname);
      break;
    case 2	:	/*executing the command	*/
      if (exec!=(char *)0) {
        char *cmd;
	int status;

        q->why=clm_freestr(q->why);
	(void) asprintf(&cmd,"cat %s | %s",mailfile,exec);
	switch (status=system(cmd)) {
	  case 0	:
	    q->status=CMDOK;
            (void) asprintf(&(q->why),"Mail piped to program <%s>",q->recname);
	    break;
	  default	:
	    q->status=REJECT;
            (void) asprintf(&(q->why),"Unable to do pipe command <%s> (status='%0x')",cmd,status);
            (void) jrl_journalling(false,EVTRT,q->why);
	    break;
 	  }
        (void) free(cmd);
	(void) free(exec);
        }
      break;
    default	:	/*SAFE Guard		*/
     proceed=false;
     break;
    }
  phase++;
  }
(void) free(mailfile);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email local quarantine	*/
/*	storage					*/
/*						*/
/************************************************/
static void localquarantine(SQLFILE *sql,QUETYP **queues)

{
char *session;
int i;

session="";
for (i=0;queues[i]!=(QUETYP *)0;i++) {
  if (strcmp(session,queues[i]->session)!=0) {
    session=queues[i]->session;
    (void) jrl_journalling(false,EVTRT,"%s %s",SESSION,queues[i]->session);
    (void) jrl_journalling(false,EVTRT,"From: <%s>",queues[i]->orgeml);
    (void) jrl_journalling(false,EVTRT,"Subject: %s",queues[i]->subject);
    }
  (void) checkdomover(sql,queues[i]);
  (void) quaremail(queues[i]);
  (void) jrl_journalling(false,EVTRT,"%s",queues[i]->why);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email piping		*/
/*						*/
/************************************************/
static void localpiping(QUETYP **queues)

{
char *session;
int i;

session="";
for (i=0;queues[i]!=(QUETYP *)0;i++) {
  if (strcmp(session,queues[i]->session)!=0) {
    session=queues[i]->session;
    (void) jrl_journalling(false,EVTRT,"%s %s",SESSION,queues[i]->session);
    (void) jrl_journalling(false,EVTRT,"From: <%s>",queues[i]->orgeml);
    (void) jrl_journalling(false,EVTRT,"Subject: %s",queues[i]->subject);
    }
  (void) piping(queues[i]);
  (void) jrl_journalling(false,EVTRT,"%s",queues[i]->why);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email local delivery	*/
/*	using the delivery shell.		*/
/*						*/
/************************************************/
static void localdelivery(SQLFILE *sql,QUETYP **queues)

{
#define	MXBXSOR	"Unable to store mail in <%s> email mailbox (status=%d), Sorry."

char *script;
char *session;
int i;

(void) asprintf(&script,"%s%s/%s",clm_appdir(d_usrlib),SHELL,DELIVER);
session="";
for (i=0;queues[i]!=(QUETYP *)0;i++) {
  char *filename;
  char *orgeml;
  char *rcpt;
  char *cmd;
  int status;

  queues[i]->why=clm_freestr(queues[i]->why);
  if (strcmp(session,queues[i]->session)!=0) {
    session=queues[i]->session;
    (void) jrl_journalling(false,EVTRT,"%s %s",SESSION,queues[i]->session);
    (void) jrl_journalling(false,EVTRT,"From: <%s>",queues[i]->orgeml);
    (void) jrl_journalling(false,EVTRT,"Subject: %s",queues[i]->subject);
    }
  (void) checkdomover(sql,queues[i]);
  filename=que_getdatafilename(d_transit,queues[i]);
  orgeml=strdup(queues[i]->orgeml);
  (void) eml_standardize(orgeml);
  (void) asprintf(&rcpt,"%s@%s",queues[i]->recname,queues[i]->recdom);
  (void) asprintf(&cmd,"%s %s \"%s\" \"%s\"",script,filename,orgeml,rcpt);
  switch (status=system(cmd)) {
    case -1		:	/*big system trouble	*/
      queues[i]->status=FAILED;
      (void) asprintf(&(queues[i]->why),"system trouble to store mail in <%s> email mailbox",rcpt);
      (void) jrl_journalling(false,EVTRT,"Bug! %s",queues[i]->why);
      break;
    case  0		:	/*everything fine	*/
      queues[i]->status=CMDOK;
      (void) asprintf(&(queues[i]->why),"mail now available in <%s> email mailbox",rcpt);
      (void) jrl_journalling(false,EVTRT,"%s",queues[i]->why);
      break;
    default		:	/*status trouble	*/	
      if (status!=0) {	/*always	*/
	const char *detail;

	detail="Unknown Error";
	queues[i]->status=CMDBAD;
        switch (WEXITSTATUS(status)) {
	  case 64	:
	    detail="Invalid parameter given";
	    break;
	  case 67	:
	    detail="The destination username was not found";
	    break;
	  case 75	:
	    detail="Temporary failure";
	    break;
	  case 77	:
	    detail="No permission to store Email";
	    break;
	  case 78	:
	    detail=" Failed to read configuration file";
	    break;
	  default	:
	    break;
          }
        (void) asprintf(&(queues[i]->why),"Unable to store mail in <%s> in mailbox (status=%d,<%s>)",
					  rcpt,WEXITSTATUS(status),detail);
        (void) jrl_journalling(false,EVTRT,"WARNING! %s",queues[i]->why);
	}
      break;
    }
  (void) free(cmd);
  (void) free(rcpt);
  (void) free(orgeml);
  (void) free(filename);
  }
(void) free(script);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare an Email to be	*/
/*	bounced back to originator.		*/
/*						*/
/************************************************/
static void rebouncing(SQLFILE *sql,QUETYP *queue,int iteration,int warning)

{
#define OPRB	"modtrt.c:rebouncing,"

char *ptr;
const char *ext;
const char *strmode;
char *destination;
MMOENU mode;
char *recipient;
char *dataname;
char *mailname;
char *rejectto;
char *bouncename;
char *subject;
char **explain;
QUETYP *qloc;
int phase;
int proceed;

ptr=(char *)0;
ext="RJCT";
strmode="Rebouncing";
if (warning==true) {
  ext="WARN";
  strmode="Warning";
  }
destination=WITHMX;
mode=mm_rebounce;
recipient=(char *)0;
dataname=(char *)0;
mailname=(char *)0;
rejectto=(char *)0;
bouncename=(char *)0;
subject=(char *)0;
explain=(char **)0;
qloc=(QUETYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*rebounce of rebounce? */
      recipient=strdup(queue->rejectto);
      queue->lang=dbc_getusrlang(sql,queue->rejectto);
      (void) asprintf(&rejectto,"%s@%s",POSTM,eml_getdomain(recipient));
      if (queue->mode==mm_rebounce) {
	ext="LASTHOPE";
	mode=mm_lasthope;
	(void) free(rejectto);
	rejectto=strdup("");
	}
      (void) asprintf(&dataname,"%s-%05lu-%03d-%s",queue->session,queue->transit,iteration,ext);
      (void) asprintf(&bouncename,"%s/%s",clm_appdir(d_transit),dataname);
      mailname=que_getdatafilename(d_transit,queue);
      subject=que_getreportsubject(queue,warning);
      break;
    case 1	:	/*preparing new queue	*/
      if ((ptr=strchr(recipient,'@'))!=(char *)0) {
	int grpid;

        grpid=queue->grpin;
	if (grpid==0) {
	  char *rcptloc;

	  (void) asprintf(&rcptloc,"%s@%s",queue->recname,queue->recdom);
	  grpid=dbc_getgroupid(sql,rcptloc);
	  (void) free(rcptloc);
	  }
        *ptr='\000';
	qloc=que_mkqueue(ANYADDR,destination,SMTPPORT,dataname,0,subject,rejectto,rejectto,
			 grpid,recipient,ptr+1,mode,(int)m_toforward);
	qloc->transit=que_getuniqtransit();
	}
      (void) free(rejectto);
      break;
    case 2	:	/*no trouble to warn or rebounce	*/
      if (qloc==(QUETYP *)0) {
	(void) clm_alert(0,"%s Unable to make rebounce to originator <%s> (queue session=<%s-%d>)",
			    OPRB,queue->orgeml,queue->session,queue->transit);
	phase=999;	/*no need to go further	*/
	}
      break;
    case 3	:	/*setting data in queue	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	QUETYP *qs[2];

	qs[0]=qloc;
        qs[1]=(QUETYP *)0;
        if (dbc_createqueues(sql,qs)==0) {
	  (void) clm_alert(0,"%s Unable to update database with queue (bug!)",OPRB);
	  qloc=que_freequeue(qloc); /*we are really in big trouble	*/
	  phase=999;
	  }
	}
      break;
    case 4	:	/*preparing attachment	*/
      queue->lang=dbc_getusrlang(sql,queue->rejectto);
      explain=que_getreportdata(queue,warning);
      if (eml_attache(mailname,bouncename,queue->session,"",queue->orgeml,subject,explain)==false) {
	(void) jrl_journalling(false,EVTRT,"%s Unable to attach original mail <%s-%d> (bug?)",
					   OPRB,queue->session,queue->transit);
        (void) unlink(bouncename);	/*back tracking	*/
	(void) trt_delqueue(sql,qloc);	/*removing queue*/
	qloc=que_freequeue(qloc);
	phase=999;		/*getting out fast	*/
	}
      explain=(char **)vec_freelstlst((void **)explain,VECFREE);
      break;
    case 5	:	/*preparing attachment	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	char *cmt;
	int ready;
	char *lfile;
	char *newshead;

	cmt="";
	ready=true;
        (void) asprintf(&newshead,"Set %s report %s to <%s@%s>",
				   strmode,qloc->session,qloc->recname,qloc->recdom);
	lfile=que_getdatafilename(d_transit,qloc);
	if (link(bouncename,lfile)<0) {
	  (void) clm_alert(0,"%s Unable to link file <%s> to file <%s> (errno=<%s>))",
			      OPRB,bouncename,lfile,strerror(errno));
	  cmt=" not queued (trouble?)";
	  ready=false;
	  }
	(void) clm_fasprintf(&newshead,"\n\t\tabout <%s@%s>",
					queue->recname,queue->recdom,cmt);
	(void) free(lfile);
	if (ready==true)
          (void) dbc_rdyqueue(sql,qloc);
	else
          (void) trt_delqueue(sql,qloc);
	(void) jrl_journalling(false,EVTRT,"%s",newshead);
	(void) free(newshead);
        (void) unlink(bouncename);
	}
      break;
    case 6	:	/*reporting the event	*/
      if (qloc!=(QUETYP *)0) {	/*always	*/
	DSPTYP dsp;
	RCPTYP rcpt;
	RCPTYP *rcpts[2];

	(void) memset(&dsp,'\000',sizeof(dsp));
	(void) memset(&rcpt,'\000',sizeof(rcpt));
	dsp.session=qloc->session;
	dsp.peerip=LOCIP4;
	dsp.peername=LOCALHOST;
	dsp.locip=LOCIP4;
	dsp.subject=qloc->subject;
	dsp.inbehalf=POSTM;
	dsp.efrom=POSTM;
	dsp.hfrom=POSTM;
	dsp.mtaille=1000;
	dsp.grpin=qloc->grpin;
	rcpt.status=m_relayed;
	rcpt.inbehalf=POSTM;
	(void) asprintf(&(rcpt.rcpt),"%s@%s",qloc->recname,qloc->recdom);
	rcpts[0]=&rcpt;
	rcpts[1]=(RCPTYP *)0;
	(void) dbc_setlogs(sql,&dsp,rcpts);
	(void) free(rcpt.rcpt);
        (void) jrl_journalling(false,EVJRL,"%-*s%s -> %s start",6,SESSION,qloc->session,strmode);
        (void) jrl_journalling(false,EVJRL,"%s: <%s>",MAILF,qloc->rejectto);
        (void) jrl_journalling(false,EVJRL,"%s: <%s@%s>",RCPT,qloc->recname,qloc->recdom);
        (void) jrl_journalling(false,EVJRL,"%s: %s","Subject",qloc->subject);
        (void) jrl_journalling(false,EVJRL,"%s end",strmode);
        (void) jrl_journalling(true,EVJRL,(char *)0);
        qloc=que_freequeue(qloc);
	}
      break;
    default	:	/*SAFE guard		*/
      queue->status=BOUNCE;	/*queue bounced	*/
      if (warning==true) 
        queue->status=RETRY;	/*queue bounced	*/
      (void) free(subject);
      (void) free(bouncename);
      mailname=clm_freestr(mailname);
      (void) free(dataname);
      (void) free(recipient);
      proceed=false;
      break;
    }
  phase++;
  }
(void) jrl_journalling(true,EVTRT,(char *)0);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transit E-mail to remote	*/
/*	SMTP server.				*/
/*						*/
/************************************************/
static void transitsmtp(SQLFILE *sql,BININF **binlst,QUETYP **queues)

{
#ifdef	DEBUG	
#define MXTIME	   15	/*15 to contact in debug*/
#define	NEXTRY     60	/*Next try in 1 minutes	*/
#else
#define MXTIME	  180	/*180s to make contact	*/
#define	NEXTRY	15*60	/*Next try in 15 minutes*/
#endif

MXTYP **mxlst;
CTTYP *contact;
int phase;
int proceed;

mxlst=(MXTYP **)0;
contact=(CTTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*lets prepare next try	*/
      if (queues!=(QUETYP **)0) {
	register time_t curtime;
	int i;

	curtime=time((time_t *)0);
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  queues[i]->lasttime=curtime;
    	  queues[i]->retry++;
	  queues[i]->nexttime=curtime+(queues[i]->retry*NEXTRY);
	  queues[i]->numtrace=que_getuniqtransit();
	  }
	}
      break;
    case 1	:		/*is E-mail relayed	*/
     (void) clm_pushdebug(5);	/*to get transit trace	*/
     if (strstr(queues[0]->smtpmx,WITHMX)!=queues[0]->smtpmx) {
       MXTYP *mx;

       mx=(MXTYP *)calloc(1,sizeof(MXTYP));
       mx->preference=1;
       mx->mxname=strdup(queues[0]->smtpmx);
       mxlst=(MXTYP **)vec_addlstlst((void *)0,(void *)mx);
       phase++;		/*no need to look for MX*/
       }
      break;
    case 2	:	/*getting serveur list	*/
      if ((mxlst=getcleanmxlst(binlst,queues[0]->recdom))==(MXTYP **)0) {
	phase++;	/*no need to do contact	*/
	phase++;	/*no need to report logs*/
	}
      break;
    case 3	:	/*looking for one server*/
      (void) jrl_journalling(false,EVTRT,"SMTP Transit to domain %s",queues[0]->recdom);
      contact=makecontact(mxlst,queues[0],MXTIME);
      break;
    case 4	:	/*reporting trouble	*/
      if (queues!=(QUETYP **)0) { 	/*always*/
	int i;

	for (i=1;queues[i]!=(QUETYP *)0;i++) {
          int j;
	  char **logs;

	  for (j=0,logs=queues[0]->logs;logs[j]!=(char *)0;j++)
  	    queues[i]->logs=(char **)vec_addlstlst((void *)queues[i]->logs,(void *)(strdup(logs[j])));
	  }
	}
      break;
    case 5	:	/*reporting trouble	*/
      if ((contact==(CTTYP *)0)||(contact->established==false)) {
	int status;
	int nummx;
        int i;

	status=INRELAY;
	nummx=vec_sizelstlst((void **)mxlst);
	if (contact==(CTTYP *)0) {
	  if (nummx>0)
	    (void) setwhy(queues,MISSMX,
			         "Unable to make SMTP contact with any of the %d MX for domain "
			         "<%s> within '%d' seconds",
		 		  nummx,queues[0]->recdom,MXTIME);
	  else {
	    status=FAILED;
	    (void) setwhy(queues,FAILED,"Unable to find even one MX for domain <%s>",queues[0]->recdom);
	    }
	  }
	else {
	  (void) setwhy(queues,FAILED,"contact with domain <%s> MX returned <%s>",queues[0]->recdom,contact->info);
	  }
	for (i=0;queues[i]!=(QUETYP *)0;i++) {
	  queues[i]->status=status;
	  (void) logqueue(sql,queues[i]);
	  }
	phase=999;	/*going out quickly	*/
	}
      break;
    case 6	:	/*remote is trusting us	*/
      if ((contact->ctxflag&VALDTED)!=0) {
	if (dbc_gettrust(sql,contact->peerip,TRSTUS)==false)
	  (void) dbc_settrust(sql,contact->peerip,TRSTUS,true,time((time_t *)0)-(H24*5));
	}
      break;
    case 7	:	/*setting rmtsrvr	*/
      if (queues!=(QUETYP **)0) { 	/*always*/
	int i;

	for (i=0;queues[i]!=(QUETYP *)0;i++) {
          queues[i]->rmtsrvr=clm_freestr(queues[i]->rmtsrvr);
          queues[i]->rmtsrvr=strdup(contact->peername);
	  }
	}
      break;
    case 8	:	/*sending Email		*/
      (void) sendemail(sql,queues,contact);
      break;
    default	:	/*SAFE guard		*/
      (void) clm_popdebug();
      contact=eml_closecontact(contact);
      mxlst=dns_cleanmx(mxlst);
      (void) jrl_journalling(true,EVTRT,(char *)0);
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to transit received Email,	*/
/*	either store the Email localy, or	*/
/*	send it to remote SMTP server		*/
/*						*/
/************************************************/
static QUETYP **dotransit(SQLFILE *sql,QUETYP **queues,int batch)

{
typedef enum	{
	a_locdel,	/*local delivery	*/
	a_locquar,	/*local quarantine	*/
	a_locpipe,	/*local piping		*/
	a_smtp		/*smtp relaying/delivery*/
	}LOCTYP;

static VOCTYP vocable[]={
        {a_locdel,LOCALDEL,(const void *)0},
        {a_locquar,LOCALQUAR,(const void *)0},
        {a_locpipe,LOCALPIPE,(const void *)0},
        {a_smtp,(char *)0,(const void *)0}
        };

register int pos;
register BININF **binlst;
register QUETYP **upd;
register QUETYP **dest;

pos=0;
binlst=dbc_getbinds(sql,(BININF **)0,pro_smtp);
binlst=dbc_getbinds(sql,binlst,pro_smtps);
upd=(QUETYP **)calloc(batch+1,sizeof(QUETYP *));
while ((dest=sortqueue(queues,srt_mx))!=(QUETYP **)0) {
  VOCTYP *voc;
  register char *smtpmx;
  register int i;

  smtpmx=strdup(dest[0]->smtpmx);
  voc=clm_getvoca(vocable,smtpmx);
  switch(voc->code) {
    case a_locdel	:/*Local delivery	*/
      (void) localdelivery(sql,dest);
      break;
    case a_locquar	:/*Local quarantine	*/
      (void) localquarantine(sql,dest);
      break;
    case a_locpipe	:/*Local piping		*/
      (void) localpiping(dest);
      break;
    case a_smtp		:/*smtp delivery/relay	*/
      (void) transitsmtp(sql,binlst,dest);
      break;
    }
  (void) free(smtpmx);
  for (i=0;dest[i]!=(QUETYP *)0;i++,pos++)
    upd[pos]=dest[i]; 
  (void) free(dest);
  }
binlst=dev_freebindinf(binlst);
(void) free(queues);
return upd;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to rebounce Email to		*/
/*	originator if Email got a reject final	*/
/*	status.					*/
/*						*/
/************************************************/
static void dorebounce(SQLFILE *sql,QUETYP **queues)

{
if (queues!=(QUETYP **)0) {
  int i;

  for (i=0;queues[i]!=(QUETYP *)0;i++) {
    if (queues[i]->status!=CMDOK) {
      if (queues[i]->upto<queues[i]->nexttime) {
	queues[i]->retry=9999;
	queues[i]->status=FAILED;
	queues[i]->why=clm_freestr(queues[i]->why);
	queues[i]->why=eml_ascstat(queues[i]->status,"Unable to transmit email in due time");
        (void) logqueue(sql,queues[i]);
	}
      if ((queues[i]->retry==TRYWARM)&&(queues[i]->mode==mm_std))
        (void) rebouncing(sql,queues[i],i,true);
      }
    if (queues[i]->status<TMPERR)
      continue;
    switch (queues[i]->mode) {
      case (mm_std)	:
      case (mm_rebounce):
	if (queues[i]->status>=SYNERR) {
	  queues[i]->nexttime=queues[i]->lasttime;
          (void) logqueue(sql,queues[i]);
          (void) rebouncing(sql,queues[i],i,false);
	  }
	else {
	  queues[i]->status=RETRY;
	  }
	break;
      default		:
	break;
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email quarantine recall	*/
/*						*/
/************************************************/
static int quarantinerecall(SQLFILE *sql,char *script,EXTTYP *ext)

{
#define	OPQU	"modtrt.c:quarantinerecall,"

int done;
char *user;
char *domain;
char *ptr;
char *filename;
char *cmd;
int status;

done=false;
(void) jrl_journalling(false,EVTRT,"Quarantined ID %s/%s",ext->session,ext->recipient);
user=strdup(ext->recipient);
domain=eml_getdomain(user);
if ((ptr=strchr(user,'@'))!=(char *)0)
  *ptr='\000';
(void) asprintf(&filename,"%s/%s/%s/%s/%s",clm_appdir(d_mails),domain,user,QUARDIR,ext->session);
(void) asprintf(&cmd,"%s \"%s\" %s \"%s@%s\"",script,filename,"postmaster",user,domain);
switch (status=system(cmd)) {
  case -1		:	/*big system trouble	*/
    (void) jrl_journalling(false,EVTRT,"Bug! system trouble to recover quarantine");
    break;
  case  0		:	/*everything fine	*/
    (void) jrl_journalling(false,EVTRT,"mail now available in <%s> email mailbox",ext->recipient);
    if (unlink(filename)<0) 
      (void) clm_alert(0,"%s Unable to remove file <%s> (error=<%s>) bug?",OPQU,filename,strerror(errno));
    (void) dbc_updlogs(sql,ext->session,ext->recipient,m_direct);
    done=true;
    break;
  default		:	/*status trouble	*/	
    (void) jrl_journalling(false,EVTRT,"WARNING! Unable to store quarantine to <%s> email mailbox (status=%d)",
					ext->recipient,WEXITSTATUS(status));
     break;
  }
(void) free(cmd);
(void) free(filename);
(void) free(user);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to do email quarantine recall	*/
/*	when the email must be sent remote	*/
/*						*/
/************************************************/
static int stayputrecall(SQLFILE *sql,EXTTYP *ext)

{
#define	OPSR	"modtrt.c:stayputrecall,"

int done;
DSPTYP *dsp;
RCPTYP **rcpts;
GRPTYP *group;
char *infile;
int phase;
int proceed;

done=false;
dsp=(DSPTYP *)0;
rcpts=(RCPTYP **)0;
group=(GRPTYP *)0;
infile=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting DSP		*/
      if ((dsp=dbc_getsessiondsp(sql,ext->session,ext->recipient))==(DSPTYP *)0){
        (void) clm_alert(0,"%s Unable to get data session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	proceed=false;
	}
      break;
    case 1	:	/*making RCPT		*/
      if (dsp!=(DSPTYP *)0) { /*always		*/
	RCPTYP *rcpt;

        rcpt=(RCPTYP *)calloc(1,sizeof(RCPTYP));
        rcpt->rcpt=strdup(ext->recipient);
        rcpts=(RCPTYP **)vec_addlstlst((void *)rcpts,(void *)rcpt);
	}
      break;
    case 2	:	/* getting group	*/
      if ((group=dbc_getgroup(sql,ext->recipient))==(GRPTYP *)0) {
        (void) clm_alert(0,"%s Unable to find group for session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	phase=999;
	}
      break;
    case 3	:	/*preparing infile	*/
      infile=eml_getquarfilename(dsp->session,ext->recipient);
      break;
    case 4	:	/*making transfer rqst	*/
      done=true;
      if (trt_transit(sql,group->exitip,group->smtprmt,group->smtpport,dsp->hfrom,dsp,infile,rcpts)==0) {
        (void) clm_alert(0,"%s Unable to recall quarantine for session=<%s> recipient=<%s> (bug?)",
			    OPSR,ext->session,ext->recipient);
	done=false;
	}
      break;
    default	:	/*SAFE Guard		*/
      infile=clm_freestr(infile);
      group=dbc_freegroup(group);
      rcpts=eml_freercpts(rcpts);
      dsp=eml_cleandispatch(dsp);
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to parse data-base banned	*/
/*	stream info, check for error and build a*/
/*	final data stream to be forwarded to	*/
/*	truting remote.				*/
/*	If the record is non parsable (DB	*/
/*	corruption), banned record is removed	*/
/*	from data-base.				*/
/*						*/
/************************************************/
static TRUTYP *checkbanlist(SQLFILE *sql,BANTYP **banlsts)

{
TRUTYP *tru;

tru=(TRUTYP *)0;
if (banlsts!=(BANTYP **)0) {
  int i;

  for (i=0;banlsts[i]!=(BANTYP *)0;i++) {
    TRUTYP *truloc;

    if ((truloc=tru_parsetrust(LOCALHOST,LOCIP4,banlsts[i]->infos))!=(TRUTYP *)0) {
      if (tru==(TRUTYP *)0) {
	tru=truloc;
	}
      else {
        tru->strms=(STRMTY **)vec_addlstlst((void **)tru->strms,(void *)truloc->strms[0]);
        truloc->strms[0]=(STRMTY *)0;
        truloc=tru_freetrusts(truloc);
	}
      }
    else {
      (void) jrl_journalling(false,APJRL,"Found info record corrupted for banned ip <%s/%d>",
					  banlsts[i]->banip,banlsts[i]->mask);
      if (banlsts[i]->infos!=(char **)0) {
	int j;
	char **ptr;

	ptr=banlsts[i]->infos;
	for (j=0;ptr[j]!=(char *)0;j++)
          (void) jrl_journalling(false,APJRL,"Info[%d]->%s",j,ptr[j]);
	}
      (void) jrl_journalling(true,APJRL,"Removed record from database");
      (void) dbc_delban(sql,banlsts[i]->banip,banlsts[i]->mask);
      }
    }
  }
return tru;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to feed trusting remote server*/
/*	with uptodate banned information.	*/
/*						*/
/************************************************/
static int doban(time_t curtime)

{
#define	OPDB	"modtrt.c:doban,"

int done;
SQLFILE *sql;
TRUDAT **trusting;
int phase;
int proceed;

done=false;
sql=(SQLFILE *)0;
trusting=(TRUDAT **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	(void) clm_alert(0,"%s Unable to re-open database",OPDB);
	done=-1;	/*error report		*/
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*getting trusting list	*/
      if ((trusting=dbc_gettrustingus(sql))!=(TRUDAT **)0) {
        int i;

	for (i=0;trusting[i]!=(TRUDAT *)0;i++) {
	  BANTYP **banlsts;
	  time_t updt;

	  updt=trusting[i]->updt;
	  if ((banlsts=dbc_getbanlist(sql,&updt))!=(BANTYP **)0) {
	    TRUTYP *tru;

	    if ((tru=checkbanlist(sql,banlsts))==(TRUTYP *)0) {
	      continue;
	      }
	    switch (trt_feedbanned(trusting[i]->ipnum,tru)) {
	      case 0	:
	        break;
	      case 1	:	/*feeding done		*/
		(void) dbc_settrust(sql,trusting[i]->ipnum,TRSTUS,true,updt);
		break;
	      case -1	:	/*no trusting anymore	*/
		(void) dbc_settrust(sql,trusting[i]->ipnum,TRSTUS,false,curtime);
		break;
	      }
            tru=tru_freetrusts(tru);
  	    banlsts=dbc_freebanlsts(banlsts);
	    }
	  }
        trusting=dbc_freetrustingus(trusting);
	done=true;
	}
      break;
    case 2	:	/*closing data base	*/
      sql=sql_close(sql);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check AND send an Email	*/
/*	to remote site				*/
/*						*/
/************************************************/
static QUETYP **tracker(QUETYP **queues)

{
#define	OPSE	"modtrt.c:tracker,"
#define	TRACKER	"%s tracker, sending %04d emails to domain <%s>"

u_int batch;
SQLFILE *sql;
int phase;
int proceed;

batch=0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database",OPSE);
	phase=999;	/*TROUBLE		*/
	}
      break;
    case 1	:	/*displaying process	*/
      batch=vec_sizelstlst((void **)queues);
      (void) prc_settitle(TRACKER,appname,(u_long)batch,queues[0]->recdom);
      (void) jrl_journalling(false,EVTRT,"domain <%s> %d email to be delivered",queues[0]->recdom,batch);
      break;
    case 2	:	/*transiting Email	*/
      queues=dotransit(sql,queues,batch);
      (void) dorebounce(sql,queues);
      break;
    case 3	:	/*final queue update	*/
      if (queues!=(QUETYP **)0) {
	int nxt;

	nxt=0;
	while (queues[nxt]!=(QUETYP *)0) {
	  (void) dbc_settrtdata(sql,queues[nxt]);
	  nxt++;
	  }
	}
      queues=trt_cleanqueues(sql,queues);
      sql=sql_close(sql);
      break;
    default	:	/*SAFE guard		*/
      (void) jrl_journalling(true,EVTRT,(char *)0);
      proceed=false;
      break;
    }
  phase++;
  }
return queues;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to scan directories extract	*/
/*	smtp file to be forwarded.		*/
/*						*/
/************************************************/
static SQLFILE *scanqueues(SQLFILE *sql,PRCTYP **prclst,SRVTYP ***srvlst,u_long *sofar)

{
#define	OPSC	"modtrt.c:scanqueues,"

#define MXQUE	200	/*maximun queue crunch size	*/
#define	MXFORK	 10	/*maximun fork number by level	*/
#define	MXSRV	  3	/*maximun chanel to one server	*/

QUETYP **queues;
u_int batch;
int letswait;
int phase;
int proceed;

queues=(QUETYP **)0;
batch=0;
letswait=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*child still working	*/
      if (allbusy(prclst,srvlst,MXFORK)==true) {
	letswait=true;
	phase=999;	/*busy, see you later	*/
	}
      break;
    case 1	:	/*something in queue?	*/
      if ((queues=dbc_getqueues(sql,MXQUE))==(QUETYP **)0) {
	letswait=true;
	phase=999;	/*no need to go further	*/
	}
      break;
    case 2	:	/*there is something	*/
      sql=sql_close(sql);
      batch=vec_sizelstlst((void **)queues);
      (void) prc_settitle(TITLE,appname,(u_long)batch,*sofar);
      if (queues!=(QUETYP **)0) {	/*always*/
        QUETYP **dest;
 
        (void) jrl_journalling(false,EVTRT,"Transitr dispatching %d emails inqueue",batch);
        while ((dest=sortqueue(queues,srt_dest))!=(QUETYP **)0) {
	  if (istoobusy(*srvlst,dest[0]->recdom,MXSRV)==true) {
	    (void) postpone(dest);
	    }
	  else {
	    GACTYP *gat;
	    pid_t transpid;

	    gat=(GACTYP *)0;
	    switch (transpid=fork()) {
	      case -1	:	/*trouble!	*/
	        (void) clm_alert(0,"%s Unable to fork transit process (error=<%s>)",OPSC,strerror(errno));
	        break;
	      case  0	:	/*child process	*/
		(void) jrl_stoppresse(EVTRT);
		(void) jrl_stoppresse(EVJRL);
    		(void) jrl_closejournals();
    		/*starting module own garbage collector	*/
    		gat=gac_open(2,appname,"tracker");
		(void) clm_setjournal(APJRL,d_log,appname);
		(void) clm_setjournal(EVJRL,d_log,"events");
		(void) clm_setjournal(EVTRT,d_log,"transit");
		if (dest!=(QUETYP **)0) {	/*always*/
		  QUETYP **clones;

		  clones=que_dupqueues(dest);
	          clones=tracker(clones);
		  clones=que_freequeues(clones);
		  }
    		(void) jrl_closejournals();
    	        (void) clm_trashmem(2,gat);
	        (void) sleep(1);
	        (void) exit(0);
	        break;
	      default	:	/*parent process*/
	        *prclst=prc_addtoprc(*prclst,transpid);
	        *srvlst=addtosrvlst(*srvlst,dest[0]->recdom,transpid);
	        (void) usleep(1000);
	        break;
	      }
	    }
          dest=que_freequeues(dest);
	  while (allbusy(prclst,srvlst,MXFORK)==true) {
	    (void) usleep(10000);
	    if (sigterm==true)
	      break;	/*do not to wait forever*/
	    }
	  }
        (void) jrl_journalling(true,EVTRT,(char *)0);
  	(void) free(queues); 
	*sofar+=batch;
        (void) prc_settitle(TITLE,appname,(u_long)0,*sofar);
	}
      break;
    case 3	:	/*reconnecting		*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to re-open database",OPSC);
	phase=999;	/*trouble trouble	*/
	}
      break;
    default	:	/*SAFE Guard		*/
      if (letswait==true) 
	(void) sleep(10);	/*lets relax	*/
      proceed=false;
      break;
    }
  phase++;
  }
return sql;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract Email to be	*/
/*	extracted from quarantine.		*/
/*						*/
/************************************************/
static void doextract(SQLFILE *sql)

{
#define	OPEX	"modtrt.c:doextract,"

EXTTYP **exts;

exts=(EXTTYP **)0;
if ((exts=dbc_getextract(sql))!=(EXTTYP **)0) {
  char *script;
  int i;

  (void) asprintf(&script,"%s%s/%s",clm_appdir(d_usrlib),SHELL,DELIVER);
  for (i=0;exts[i]!=(EXTTYP *)0;i++) {
    LOGTYP *emlog;
    int done;

    emlog=(LOGTYP *)0;
    done=false;
    switch (exts[i]->status) {
      case m_quarantine	:
        done=quarantinerecall(sql,script,exts[i]);
	break;
      case m_stayput	:
        done=stayputrecall(sql,exts[i]);
	break;
      default		:
	(void) clm_alert(0,"%s Unexpected status '%d' for quarantine recall session=<%s> recipient=<%s> (Bug?)",
			   "modtrt.c:doextract,",exts[i]->status,exts[i]->session,exts[i]->recipient);
	break;
      }
    if (done==false)
      continue;
    if ((emlog=dbc_getlogs(sql,exts[i]->session,exts[i]->recipient))!=(LOGTYP *)0) {
      char *orgdom;

      orgdom=eml_getdomain(emlog->efrom);
      if (dbc_setoverride(sql,emlog->recipient,"*",orgdom,emlog->rmtip,16,tr_accepted)==false) {
        (void) clm_alert(0,"%s Unable to set override recipient <%s> origin=<%s> rmtip=<%s> (bug?)",
			   OPEX,emlog->recipient,emlog->hfrom,emlog->rmtip);
	}
      emlog=dbc_freelogs(emlog);
      }
    }
  (void) free(script);
  exts=dbc_freeextract(exts);
  (void) jrl_journalling(true,EVTRT,(char *)0);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to start a unique process to	*/
/*	send to trusting remote banned data to	*/
/*	exchange.				*/
/*						*/
/************************************************/
static int startdoban(SQLFILE **sql,time_t curtime)

{
#define	OPSB	"modtrt.c:startdoban,"

static	pid_t dobanpid=(pid_t)0;
int done;
GACTYP *gat;
int phase;
int proceed;

done=false;
gat=(GACTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*doban still pending?	*/
      (void) prc_nozombie();
      if (prc_checkprocess(dobanpid)==true)
        phase=999;	/*process still pending!*/
      break;
    case 1	:	/*preparing fork	*/
      *sql=sql_close(*sql);
      switch (dobanpid=fork()) {
	case -1	:	/*trouble!		*/
	  (void) clm_alert(0,"%s Unable to fork transit process (error=<%s>)",OPSB,strerror(errno));
	  break;
	case  0	:	/*child process		*/
          (void) jrl_stoppresse(EVTRT);
	  (void) jrl_stoppresse(EVJRL);
	  (void) jrl_closejournals();
    	  /*module own garbage collector	*/
    	  gat=gac_open(2,appname,"banned");
	  (void) clm_setjournal(APJRL,d_log,appname);
	  (void) clm_setjournal(EVTRU,d_log,"trusting");
	  (void) doban(curtime);
	  (void) jrl_closejournals();
    	  (void) clm_trashmem(2,gat);
	  (void) sleep(1);
	  (void) exit(0);
	  break;
        default	:	/*parent process	*/
	  done=true;
	  (void) usleep(1000);
	  break;
	}
      break;
    case 2	:	/*opening back data-base*/
      if ((*sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
	done=-1;	/*error report		*/
	phase=999;	/*trouble trouble	*/
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to wait for email transit	*/
/*	order.					*/
/*						*/
/************************************************/
static void scanner()

{
#define OPSA	"modtrt.c:scanner,"

#define	QUATIM	5	/*quarantine wakeup time*/
#define	BANTIM	60	/*ban list wakeup time	*/

time_t lastclean;
time_t lastquar;
time_t lastban;
u_long sofar;
PRCTYP *prclst;
SRVTYP **srvlst;
SQLFILE *sql;
int phase;
int proceed;

lastclean=(time_t)0;
lastquar=(time_t)0;
lastban=(time_t)0;
sofar=(u_long)0;
prclst=(PRCTYP *)0;
srvlst=(SRVTYP **)0;
sql=(SQLFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  time_t curtime;

  (void) prc_nozombie();
  curtime=time((time_t *)0);
  switch (phase) {
    case 0	:	/*openning data-base	*/
      if ((sql=sql_open(empty,empty,empty))==(SQLFILE *)0) {
        (void) clm_alert(0,"%s Unable to open database",OPSA);
	phase=999;	/*trouble trouble	*/
	break;
	}
      (void) dbc_updlostqueue(sql);
      break;
    case 1	:	/*main loop, cleaning	*/
      if ((lastclean+60)<curtime) {
	(void) trt_dbcleanup(sql,curtime);
	lastclean=curtime;
	}
      break;
    case 2	:	/*check quarantine reqst*/
      if ((lastquar+QUATIM)<curtime) {
        (void) doextract(sql);
	lastquar=curtime;
	}
      break;
    case 3	:	/*time for banned list	*/
      if ((lastban+BANTIM)<curtime) {
	switch (startdoban(&sql,curtime)) {
	  case false :
	    break;
	  case true  :
	    lastban=curtime;
	    break;
	  default    :	/*SQL lost!		*/
            (void) clm_alert(0,"%s Trouble! scanner exit",OPSA);
	    phase=999;	/*trouble		*/
	    break;
	  }
	}
      break;
    case 4	:	/*must be last phase	*/
      if ((sql=scanqueues(sql,&prclst,&srvlst,&sofar))!=(SQLFILE *)0)
	phase=0;	/*lets go to phase 1	*/
      break;
    default	:	/*SAFE Guard		*/
      prclst=prc_freeprc(prclst);
      srvlst=chksrvlst(prclst,srvlst);
      sql=sql_close(sql);
      proceed=false;
      break;
    }
  phase++;
  if (sigterm==true) 
    phase=999;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to fork a transit process, 	*/
/*	this process whill handle going-out	*/
/*	SMTP queue.				*/
/*						*/
/************************************************/
pid_t trt_cycle()


{
#define	OPTR	"modtrt.c:trt_cycle,"

pid_t trt;
GACTYP *gat;

gat=(GACTYP *)0;
switch (trt=fork()) {
  case -1	:
    (void) clm_alert(0,"%s unable to fork (error=<%s>)",OPTR,strerror(errno));
    break;
  case  0	:
    (void) jrl_closejournals();
    /*starting module own garbage collector	*/
    gat=gac_open(2,appname,"transiter");
    (void) clm_setjournal(APJRL,d_log,appname);
    (void) clm_setjournal(EVJRL,d_log,"events");
    (void) clm_setjournal(EVTRT,d_log,"transit");
    (void) prc_settitle(TITLE,appname,(u_long)0,(u_long)0);
    (void) scanner();
    (void) jrl_closejournals();
    (void) clm_trashmem(2,gat);
    (void) sleep(1);
    (void) exit(0);
    break;
  default	:
    break;
  }
return trt;
}
